@@ -12,6 +12,7 @@ from typing import Dict, List
 from joblib import Parallel, delayed, parallel_backend
 from jsonschema import Draft7Validator, FormatChecker
 from singer import get_logger
+from datetime import datetime, timedelta
 
 import target_snowflake.file_formats.csv as csv
 import target_snowflake.file_formats.parquet as parquet
@@ -60,22 +61,47 @@ def emit_state(state):
         sys.stdout.flush()
 
 
-def load_table_cache(config):
-    """Load table cache from snowflake metadata"""
+def get_snowflake_statics(config):
+    """Retrieve common Snowflake items will be used multiple times
+
+    Params:
+        config: configuration dictionary
+
+    Returns:
+        tuple of retrieved items: table_cache, file_format_type
+    """
     table_cache = []
     if not ('disable_table_cache' in config and config['disable_table_cache']):
         LOGGER.info('Getting catalog objects from table cache...')
 
-        db = DbSync(config) # pylint: disable=invalid-name
+        db = DbSync(config)  # pylint: disable=invalid-name
         table_cache = db.get_table_columns(
             table_schemas=stream_utils.get_schema_names_from_config(config))
 
-    return table_cache
+    # The file format is detected at DbSync init time
+    file_format_type = db.file_format.file_format_type
+
+    return table_cache, file_format_type
 
 
 # pylint: disable=too-many-locals,too-many-branches,too-many-statements,invalid-name
-def persist_lines(config, lines, table_cache=None) -> None:
-    """Main loop to read and consume singer messages from stdin"""
+def persist_lines(config, lines, table_cache=None, file_format_type: FileFormatTypes = None) -> None:
+    """Main loop to read and consume singer messages from stdin
+
+    Params:
+        config: configuration dictionary
+        lines: iterable of singer messages
+        table_cache: Optional dictionary of Snowflake table structures. This is useful to run the less
+                     INFORMATION_SCHEMA and SHOW queries as possible.
+                     If not provided then an SQL query will be generated at runtime to
+                     get all the required information from Snowflake
+        file_format_type: Optional FileFormatTypes value that defines which supported file format to use
+                          to load data into Snowflake.
+                          If not provided then it will be detected automatically
+
+    Returns:
+        tuple of retrieved items: table_cache, file_format_type
+    """
     state = None
     flushed_state = None
     schemas = {}
@@ -86,6 +112,10 @@ def persist_lines(config, lines, table_cache=None) -> None:
     stream_to_sync = {}
     total_row_count = {}
     batch_size_rows = config.get('batch_size_rows', DEFAULT_BATCH_SIZE_ROWS)
+    batch_wait_limit_seconds = config.get('batch_wait_limit_seconds', None)
+    flush_timestamp = datetime.utcnow()
+    archive_load_files = config.get('archive_load_files', False)
+    archive_load_files_data = {}
 
     # Loop over lines from stdin
     for line in lines:
@@ -143,7 +173,33 @@ def persist_lines(config, lines, table_cache=None) -> None:
             else:
                 records_to_load[stream][primary_key_string] = o['record']
 
+            if archive_load_files and stream in archive_load_files_data:
+                # Keep track of min and max of the designated column
+                stream_archive_load_files_values = archive_load_files_data[stream]
+                if 'column' in stream_archive_load_files_values:
+                    incremental_key_column_name = stream_archive_load_files_values['column']
+                    incremental_key_value = o['record'][incremental_key_column_name]
+                    min_value = stream_archive_load_files_values['min']
+                    max_value = stream_archive_load_files_values['max']
+
+                    if min_value is None or min_value > incremental_key_value:
+                        stream_archive_load_files_values['min'] = incremental_key_value
+
+                    if max_value is None or max_value < incremental_key_value:
+                        stream_archive_load_files_values['max'] = incremental_key_value
+
+            flush = False
             if row_count[stream] >= batch_size_rows:
+                flush = True
+                LOGGER.info("Flush triggered by batch_size_rows (%s) reached in %s",
+                            batch_size_rows, stream)
+            elif (batch_wait_limit_seconds and
+                  datetime.utcnow() >= (flush_timestamp + timedelta(seconds=batch_wait_limit_seconds))):
+                flush = True
+                LOGGER.info("Flush triggered by batch_wait_limit_seconds (%s)",
+                            batch_wait_limit_seconds)
+
+            if flush:
                 # flush all streams, delete records if needed, reset counts and then emit current state
                 if config.get('flush_all_streams'):
                     filter_streams = None
@@ -158,8 +214,11 @@ def persist_lines(config, lines, table_cache=None) -> None:
                     config,
                     state,
                     flushed_state,
+                    archive_load_files_data,
                     filter_streams=filter_streams)
 
+                flush_timestamp = datetime.utcnow()
+
                 # emit last encountered state
                 emit_state(copy.deepcopy(flushed_state))
 
@@ -186,7 +245,8 @@ def persist_lines(config, lines, table_cache=None) -> None:
                                                   stream_to_sync,
                                                   config,
                                                   state,
-                                                  flushed_state)
+                                                  flushed_state,
+                                                  archive_load_files_data)
 
                     # emit latest encountered state
                     emit_state(flushed_state)
@@ -210,9 +270,33 @@ def persist_lines(config, lines, table_cache=None) -> None:
                 key_properties[stream] = o['key_properties']
 
                 if config.get('add_metadata_columns') or config.get('hard_delete'):
-                    stream_to_sync[stream] = DbSync(config, add_metadata_columns_to_schema(o), table_cache)
+                    stream_to_sync[stream] = DbSync(config,
+                                                    add_metadata_columns_to_schema(o),
+                                                    table_cache,
+                                                    file_format_type)
                 else:
-                    stream_to_sync[stream] = DbSync(config, o, table_cache)
+                    stream_to_sync[stream] = DbSync(config, o, table_cache, file_format_type)
+
+                if archive_load_files:
+                    archive_load_files_data[stream] = {
+                        'tap': config.get('tap_id'),
+                    }
+
+                    # In case of incremental replication, track min/max of the replication key.
+                    # Incremental replication is assumed if o['bookmark_properties'][0] is one of the columns.
+                    incremental_key_column_name = stream_utils.get_incremental_key(o)
+                    if incremental_key_column_name:
+                        LOGGER.info("Using %s as incremental_key_column_name", incremental_key_column_name)
+                        archive_load_files_data[stream].update(
+                            column=incremental_key_column_name,
+                            min=None,
+                            max=None
+                        )
+                    else:
+                        LOGGER.warning(
+                            "archive_load_files is enabled, but no incremental_key_column_name was found. "
+                            "Min/max values will not be added to metadata for stream %s.", stream
+                        )
 
                 stream_to_sync[stream].create_schema_if_not_exists()
                 stream_to_sync[stream].sync_table()
@@ -238,7 +322,8 @@ def persist_lines(config, lines, table_cache=None) -> None:
     # then flush all buckets.
     if sum(row_count.values()) > 0:
         # flush all streams one last time, delete records if needed, reset counts and then emit current state
-        flushed_state = flush_streams(records_to_load, row_count, stream_to_sync, config, state, flushed_state)
+        flushed_state = flush_streams(records_to_load, row_count, stream_to_sync, config, state, flushed_state,
+                                      archive_load_files_data)
 
     # emit latest state
     emit_state(copy.deepcopy(flushed_state))
@@ -252,6 +337,7 @@ def flush_streams(
         config,
         state,
         flushed_state,
+        archive_load_files_data,
         filter_streams=None):
     """
     Flushes all buckets and resets records count to 0 as well as empties records to load list
@@ -262,6 +348,7 @@ def flush_streams(
     :param state: dictionary containing the original state from tap
     :param flushed_state: dictionary containing updated states only when streams got flushed
     :param filter_streams: Keys of streams to flush from the streams dict. Default is every stream
+    :param archive_load_files_data: dictionary of dictionaries containing archive load files data
     :return: State dict with flushed positions
     """
     parallelism = config.get("parallelism", DEFAULT_PARALLELISM)
@@ -294,7 +381,8 @@ def flush_streams(
             db_sync=stream_to_sync[stream],
             no_compression=config.get('no_compression'),
             delete_rows=config.get('hard_delete'),
-            temp_dir=config.get('temp_dir')
+            temp_dir=config.get('temp_dir'),
+            archive_load_files=copy.copy(archive_load_files_data.get(stream, None))
         ) for stream in streams_to_flush)
 
     # reset flushed stream records to empty to avoid flushing same records
@@ -315,16 +403,20 @@ def flush_streams(
         else:
             flushed_state = copy.deepcopy(state)
 
+        if stream in archive_load_files_data:
+            archive_load_files_data[stream]['min'] = None
+            archive_load_files_data[stream]['max'] = None
+
     # Return with state message with flushed positions
     return flushed_state
 
 
 def load_stream_batch(stream, records, row_count, db_sync, no_compression=False, delete_rows=False,
-                      temp_dir=None):
+                      temp_dir=None, archive_load_files=None):
     """Load one batch of the stream into target table"""
     # Load into snowflake
     if row_count[stream] > 0:
-        flush_records(stream, records, db_sync, temp_dir, no_compression)
+        flush_records(stream, records, db_sync, temp_dir, no_compression, archive_load_files)
 
         # Delete soft-deleted, flagged rows - where _sdc_deleted at is not null
         if delete_rows:
@@ -338,7 +430,8 @@ def flush_records(stream: str,
                   records: List[Dict],
                   db_sync: DbSync,
                   temp_dir: str = None,
-                  no_compression: bool = False) -> None:
+                  no_compression: bool = False,
+                  archive_load_files: Dict = None) -> None:
     """
     Takes a list of record messages and loads it into the snowflake target table
 
@@ -348,8 +441,9 @@ def flush_records(stream: str,
                  column value
         row_count:
         db_sync: A DbSync object
-        temp_dir: Directory where intermediate temporary files will be created. (Default: OS specificy temp directory)
+        temp_dir: Directory where intermediate temporary files will be created. (Default: OS specific temp directory)
         no_compression: Disable to use compressed files. (Default: False)
+        archive_load_files: Data needed for archive load files. (Default: None)
 
     Returns:
         None
@@ -360,7 +454,7 @@ def flush_records(stream: str,
                                                              compression=not no_compression,
                                                              dest_dir=temp_dir,
                                                              data_flattening_max_level=
-                                                                db_sync.data_flattening_max_level)
+                                                             db_sync.data_flattening_max_level)
 
     # Get file stats
     row_count = len(records)
@@ -370,8 +464,39 @@ def flush_records(stream: str,
     s3_key = db_sync.put_to_stage(filepath, stream, row_count, temp_dir=temp_dir)
     db_sync.load_file(s3_key, row_count, size_bytes)
 
-    # Delete file from local disk and from s3
+    # Delete file from local disk
     os.remove(filepath)
+
+    if archive_load_files:
+        stream_name_parts = stream_utils.stream_name_to_dict(stream)
+        if 'schema_name' not in stream_name_parts or 'table_name' not in stream_name_parts:
+            raise Exception("Failed to extract schema and table names from stream '{}'".format(stream))
+
+        archive_schema = stream_name_parts['schema_name']
+        archive_table = stream_name_parts['table_name']
+        archive_tap = archive_load_files['tap']
+
+        archive_metadata = {
+            'tap': archive_tap,
+            'schema': archive_schema,
+            'table': archive_table,
+            'archived-by': 'pipelinewise_target_snowflake'
+        }
+
+        if 'column' in archive_load_files:
+            archive_metadata.update({
+                'incremental-key': archive_load_files['column'],
+                'incremental-key-min': str(archive_load_files['min']),
+                'incremental-key-max': str(archive_load_files['max'])
+            })
+
+        # Use same file name as in import
+        archive_file = os.path.basename(s3_key)
+        archive_key = "{}/{}/{}".format(archive_tap, archive_table, archive_file)
+
+        db_sync.copy_to_archive(s3_key, archive_key, archive_metadata)
+
+    # Delete file from S3
     db_sync.delete_from_stage(stream, s3_key)
 
 
@@ -388,11 +513,11 @@ def main():
         config = {}
 
     # Init columns cache
-    table_cache = load_table_cache(config)
+    table_cache, file_format_type = get_snowflake_statics(config)
 
     # Consume singer messages
     singer_messages = io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8')
-    persist_lines(config, singer_messages, table_cache)
+    persist_lines(config, singer_messages, table_cache, file_format_type)
 
     LOGGER.debug("Exiting normally")
 
